# Copyright (c) 2010 gocept gmbh & co. kg
# See also LICENSE.txt

import ConfigParser
import datetime
import gocept.support
import gocept.support.utils
import gocept.support.notification
import gocept.support.staff
import gocept.support.rt
import icalendar
import optparse
import os
import pytz
import sys


def delta_to_hours(delta):
    return delta.days * 24 + delta.seconds/3600


class Period(object):
    """A specific period during which support needs to be provided."""

    def __init__(self, start, end, description, response_time, notify_upcoming):
        self.start = start
        self.end = end
        self.description = description
        self.response_time = response_time
        self.notify_upcoming = notify_upcoming


class SLA(object):

    def __init__(self, filename):
        self.cfg = ConfigParser.ConfigParser()
        self.cfg.read(filename)
        self.customer = self.cfg.get('general', 'customer')
        self.pin = self.cfg.get('general', 'pin')
        self.description = self.cfg.get('general', 'description')
        self.tz = pytz.timezone(self.cfg.get('general', 'timezone'))

    @classmethod
    def from_pin(cls, conf, pin):
        for sla, sla_file in conf.items('slas'):
            sla = cls(sla_file)
            if sla.pin == pin:
                return sla
        raise KeyError()

    def _get(self, section, key, default=None):
        try:
            value = self.cfg.get(section, key)
        except ConfigParser.NoOptionError:
            value = default
        return value

    def _time_for_day(self, time_str, day):
        if time_str == '24:00':
            time_str = None
        if time_str is None:
            return
        time = datetime.datetime.strptime(time_str, '%H:%M')
        t = time.replace(day=day.day, month=day.month, year=day.year)
        t = self.tz.localize(t)
        return t

    def period(self, time):
        """Return the period active at the given time."""
        for period in self.periods(time.date()):
            if period.start is not None and period.start > time:
                continue
            if period.end is not None and period.end < time:
                continue
            return period
        raise KeyError('No supported booked at %r' % time)

    def next_period(self, time):
        """Search for the next active period for _time_."""
        # does time match the currently active period?
        try:
            return self.period(time)
        except KeyError:
            pass
        # search upcoming periods
        attempt = 0
        d = time.date()
        while attempt < 180:
            for period in self.periods(d):
                if period.start is not None and period.end >= time:
                    return period
            d += datetime.timedelta(days=1)
            attempt += 1

    def periods(self, day):
        """Return period data for the given day."""
        # Determine which periods apply for the given day
        try:
            periods = self.cfg.get('days', day.isoformat())
        except ConfigParser.NoOptionError:
            try:
                periods = self.cfg.get('days', day.strftime('%A'))
            except ConfigParser.NoOptionError:
                periods = ''

        periods = [x.strip() for x in periods.split(',') if x]

        for period in periods:
            response_time = int(self.cfg.get(period, 'response_time'))

            notify_upcoming = self._get(period, 'notify_upcoming', True)
            if notify_upcoming is not None:
                notify_upcoming = bool(int(notify_upcoming))

            start = self._get(period, 'start')
            start = self._time_for_day(start, day)

            end = self._get(period, 'end')
            end = self._time_for_day(end, day)

            yield Period(start, end, period, response_time, notify_upcoming)

    def response_deadline(self, time):
        """Determine the time by which a response to a request should be given.

        If time does not lie within the current service period, the remaining
        time is stretched over the next service periods.
        """
        p = self.next_period(time)
        time_left = datetime.timedelta(minutes=p.response_time)
        period_start = max([time, p.start])
        while period_start + time_left > p.end:
            time_left -= p.end - period_start
            p = self.next_period(p.end + datetime.timedelta(seconds=1))
            period_start = p.start
        return period_start + time_left

    def events(self, begin=None, end=None):
        if begin is None:
            begin = datetime.date.today() - datetime.timedelta(days=60)
        if end is None:
            end = datetime.date.today() + datetime.timedelta(days=180)

        current = begin
        while current <= end:
            for period in self.periods(current):
                event = icalendar.Event()
                event.add('summary', '%s (%s): %s (%s minutes)' %
                          (self.cfg.get('general', 'description'),
                           self.cfg.get('general', 'customer'),
                           period.description,
                           period.response_time))

                pstart = period.start
                pend = period.end
                if pstart is None and pend is None:
                    pstart = pend = current
                elif pstart is None:
                    pstart = self._time_for_day('00:00', current)
                elif pend is None:
                    pend = self._time_for_day('23:59', current)

                event.add('dtstart', pstart)
                event.add('dtend', pend)
                event.add('dtstamp', datetime.datetime.now(self.tz))
                yield event
            current += datetime.timedelta(days=1)


def create_ical_for_slas():
    parser = optparse.OptionParser()
    parser.add_option('-o', '--output', default=None)
    parser.add_option('-c', '--config', default='~/.gocept-support.cfg')
    options, args = parser.parse_args()
    conf = ConfigParser.ConfigParser()
    conf.read([os.path.expanduser(options.config)])

    cal = icalendar.Calendar()
    cal.add('prodid', '-//gocept.support//gocept//')

    for sla, sla_file in conf.items('slas'):
        sla = SLA(sla_file)
        for event in sla.events():
            cal.add_component(event)

    if options.output is None:
        f = sys.stdout
    else:
        f = open(options.output, 'w')
    f.write(cal.as_string())


def notify_upcoming_periods():
    "Create a ticket and notify supports about an upcoming support period."""
    parser = optparse.OptionParser()
    parser.add_option('-o', '--output', default=None)
    parser.add_option('-c', '--config', default='~/.gocept-support.cfg')
    options, args = parser.parse_args()
    conf = ConfigParser.ConfigParser()
    conf.read([os.path.expanduser(options.config)])

    rt = gocept.support.rt.RT(conf.get('rt', 'user'),
                              conf.get('rt', 'password'),
                              conf.get('rt', 'url'))

    notifications = gocept.support.notification.Notifications(conf)

    for sla, sla_file in conf.items('slas'):
        sla = SLA(sla_file)
        now = sla.tz.localize(datetime.datetime.now())
        for day in [datetime.date.today(),
                    datetime.date.today()+datetime.timedelta(days=1)]:
            for period in sla.periods(day):
                if not period.notify_upcoming:
                    continue
                pstart = period.start
                if pstart is None:
                    pstart = sla._time_for_day('00:00', day)
                if period.response_time > 45:
                    horizon_decrease = 15
                else:
                    horizon_decrease = 60 - period.response_time
                horizon = pstart - datetime.timedelta(minutes=horizon_decrease)
                if now < horizon:
                    continue
                if now > pstart:
                    continue

                subject = '%s (%smin) on %s-%s' % (
                    sla.description, period.response_time,
                    period.start.strftime('%b%d %H:%M'),
                    period.end.strftime('%H:%M'))

                # Try to find an existing ticket, in case we were already run.
                tickets = rt.search("Subject = '%s'" % subject)
                if tickets:
                    continue

                try:
                    ticket = rt.new_ticket({'Queue': conf.get('rt', 'queue'),
                                   'Subject': subject})
                    due = period.start + datetime.timedelta(
                        minutes=period.response_time)
                    due = due.astimezone(pytz.utc).strftime(rt.DATE_FORMAT)
                    ticket.edit({'Starts': due})
                except:
                    ticket = dict(id='n/a')

                notifications.notify('#%s %s' % (ticket['id'], subject), 'Please prepare.')


def check_sla_tickets():
    """This test checks that no SLA tickets are running close to their
    deadline."""
    parser = optparse.OptionParser()
    parser.add_option('-c', '--config', default='~/.gocept-support.cfg')
    options, args = parser.parse_args()
    conf = ConfigParser.ConfigParser()
    conf.read([os.path.expanduser(options.config)])

    warning = []
    critical = []

    rt = gocept.support.rt.RT(conf.get('rt', 'user'),
                              conf.get('rt', 'password'),
                              conf.get('rt', 'url'))
    tickets = rt.search("Queue='%s' AND Status = 'new'" %
                        conf.get('rt', 'queue'))

    now = datetime.datetime.now()
    warning_deadline = now + datetime.timedelta(minutes=60)
    critical_deadline = now + datetime.timedelta(minutes=15)

    # Go through all new tickets and check their SLA deadline
    for ticket in tickets:
        if ticket['Starts'] == 'Not set':
            continue
        deadline = datetime.datetime.strptime(
            ticket['Starts'], rt.DATE_FORMAT)
        if deadline < critical_deadline:
            critical.append(ticket)
            continue
        if deadline < warning_deadline:
            warning.append(ticket)

    if critical:
        print "Due in < 15 minutes: %s" % (
            ', '.join('#'+x['id'] for x in critical))
        sys.exit(2)
    if warning:
        print "Due in < 60 minutes: %s" % (
            ', '.join('#'+x['id'] for x in warning))
        sys.exit(1)

    print "%s new tickets, no short term deadlines" % len(tickets)
    sys.exit(0)


def set_response_deadlines():
    parser = optparse.OptionParser()
    parser.usage = '%progname [options] working_hours_sla'
    parser.add_option('-c', '--config', default='~/.gocept-support.cfg',
        help='path to master config file')
    options, args = parser.parse_args()
    conf = ConfigParser.ConfigParser()
    conf.read([os.path.expanduser(options.config)])
    if len(args) != 1:
        parser.print_help();
        sys.exit(1)
    sla = SLA(conf.get('slas', args[0]))
    rt = gocept.support.rt.RT(conf.get('rt', 'user'),
                              conf.get('rt', 'password'),
                              conf.get('rt', 'url'))
    tickets = rt.search("Queue='%s' AND Status = 'new'" %
                        conf.get('rt', 'queue'))
    for ticket in tickets:
        if ticket['Starts'] != 'Not set':
            continue
        created = sla.tz.localize(datetime.datetime.strptime(ticket['Created'],
            rt.DATE_FORMAT))
        starts = sla.response_deadline(created)
        ticket.edit({'Starts':
                     starts.astimezone(pytz.utc).strftime(rt.DATE_FORMAT)})


def sla_report():
    parser = optparse.OptionParser()
    parser.usage = '%progname [options] sla month'
    parser.add_option('-c', '--config', default='~/.gocept-support.cfg',
        help='path to master config file')
    options, args = parser.parse_args()
    conf = ConfigParser.ConfigParser()
    conf.read([os.path.expanduser(options.config)])
    if len(args) != 2:
        parser.print_help();
        sys.exit(1)
    sla = SLA(conf.get('slas', args[0]))
    year, month = args[1].split('-')
    first = datetime.date(year=int(year), month=int(month, 10), day=1)

    def iter_month_periods():
        current = first
        while current.month == first.month:
            for period in sla.periods(current):
                yield period.description, current, delta_to_hours(period.end-period.start)
            current += datetime.timedelta(days=1)

    totals = {}
    by_day = {}
    for period, day, duration in iter_month_periods():
        totals.setdefault(period, 0)
        totals[period] += duration
        by_day.setdefault(day, {})
        by_day[day][period] = duration

    print gocept.support.utils.templates['sla_report.pt'](
        month=args[1],
        by_day=by_day,
        totals=totals).encode('utf-8')
